# frozen_string_literal: true

# Released under the MIT License.
# Copyright, 2011-2018, by Tony Arcieri.
# Copyright, 2018-2024, by Samuel Williams.
# Copyright, 2023, by Tsimnuj Hawj.

module NIO
  VERSION = "2.7.5"
end
